@@ -4,13 +4,14 @@ use std::ptr;
 use gdk_pixbuf::{Colorspace, Pixbuf};
 use gio::prelude::*;
 use glib::translate::*;
-use rgb::{FromSlice, RGBA8};
+use rgb::{AsPixels, RGBA8};
 use url::Url;
 
 use crate::c_api::checked_i32;
 
 use rsvg_internals::{
-    Dpi, Handle, LoadOptions, LoadingError, Pixel, RenderingError, SharedImageSurface, SurfaceType,
+    Dpi, Handle, LoadOptions, LoadingError, Pixel, RenderingError, SharedImageSurface,
+    SurfaceType,
 };
 
 use crate::c_api::set_gerror;
@@ -53,7 +54,7 @@ pub fn pixbuf_from_surface(surface: &SharedImageSurface) -> Result<Pixbuf, Rende
     let pixbuf_rows = pixels.chunks_mut(stride).take(height);
 
     for (src_row, dest_row) in surface.rows().zip(pixbuf_rows) {
-        let row: &mut [RGBA8] = dest_row[..width_in_bytes].as_rgba_mut();
+        let row: &mut [RGBA8] = dest_row[..width_in_bytes].as_pixels_mut();
 
         for (src, dest) in src_row.iter().zip(row.iter_mut()) {
             let (r, g, b, a) = if src.a == 0 {
@@ -64,8 +65,7 @@ pub fn pixbuf_from_surface(surface: &SharedImageSurface) -> Result<Pixbuf, Rende
                     g: src.g,
                     b: src.b,
                     a: src.a,
-                }
-                .unpremultiply();
+                }.unpremultiply();
 
                 (pixel.r, pixel.g, pixel.b, pixel.a)
             };
