@@ -1,17 +1,21 @@
 //! Gradient paint servers; the `linearGradient` and `radialGradient` elements.
 
 use cssparser::Parser;
-use markup5ever::{expanded_name, local_name, namespace_url, ns};
+use markup5ever::{
+    expanded_name, local_name, namespace_url, ns, ExpandedName, LocalName, Namespace,
+};
+use matches::matches;
 use std::cell::RefCell;
 
 use crate::allowed_url::Fragment;
 use crate::bbox::*;
 use crate::coord_units::CoordUnits;
-use crate::document::{AcquiredNode, AcquiredNodes, NodeStack};
+use crate::document::{AcquiredNodes, NodeStack};
 use crate::drawing_ctx::{DrawingCtx, ViewParams};
+use crate::element::{Draw, Element, ElementResult, SetAttributes};
 use crate::error::*;
 use crate::length::*;
-use crate::node::{CascadedValues, NodeResult, NodeTrait, NodeType, RsvgNode};
+use crate::node::{CascadedValues, Node, NodeBorrow};
 use crate::paint_server::{AsPaintSource, PaintSource};
 use crate::parsers::{Parse, ParseValue};
 use crate::properties::ComputedValues;
@@ -33,7 +37,7 @@ struct ColorStop {
     opacity: UnitInterval,
 }
 
-// gradientUnits attribute; its default is objectBoundingBox
+// gradientUnits attibute; its default is objectBoundingBox
 coord_units!(GradientUnits, CoordUnits::ObjectBoundingBox);
 
 /// spreadMethod attribute for gradients
@@ -71,45 +75,6 @@ impl From<SpreadMethod> for cairo::Extend {
     }
 }
 
-// SVG defines radial gradients as being inside a circle (cx, cy, radius).  The
-// gradient projects out from a focus point (fx, fy), which is assumed to be
-// inside the circle, to the edge of the circle.
-// The description of https://www.w3.org/TR/SVG/pservers.html#RadialGradientElement
-// states:
-//
-// If the point defined by ‘fx’ and ‘fy’ lies outside the circle defined by
-// ‘cx’, ‘cy’ and ‘r’, then the user agent shall set the focal point to the
-// intersection of the line from (‘cx’, ‘cy’) to (‘fx’, ‘fy’) with the circle
-// defined by ‘cx’, ‘cy’ and ‘r’.
-//
-// So, let's do that!
-fn fix_focus_point(fx: f64, fy: f64, cx: f64, cy: f64, radius: f64) -> (f64, f64) {
-    // Easy case first: the focus point is inside the circle
-
-    if (fx - cx) * (fx - cx) + (fy - cy) * (fy - cy) <= radius * radius {
-        return (fx, fy);
-    }
-
-    // Hard case: focus point is outside the circle.
-    // Find the vector from the origin to (fx, fy)
-
-    let mut dx = fx - cx;
-    let mut dy = fy - cy;
-
-    // Find the vector's magnitude
-    let mag = (dx * dx + dy * dy).sqrt();
-
-    // Normalize the vector to have a magnitude equal to radius
-    let scale = mag / radius;
-
-    dx /= scale;
-    dy /= scale;
-
-    // Translate back to (cx, cy) and we are done!
-
-    (cx + dx, cy + dy)
-}
-
 /// Node for the <stop> element
 #[derive(Default)]
 pub struct Stop {
@@ -128,8 +93,8 @@ fn validate_offset(length: Length<Both>) -> Result<Length<Both>, ValueErrorKind>
     }
 }
 
-impl NodeTrait for Stop {
-    fn set_atts(&mut self, _: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
+impl SetAttributes for Stop {
+    fn set_attributes(&mut self, pbag: &PropertyBag<'_>) -> ElementResult {
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
                 expanded_name!("", "offset") => {
@@ -145,6 +110,8 @@ impl NodeTrait for Stop {
     }
 }
 
+impl Draw for Stop {}
+
 /// Parameters specific to each gradient type, before being resolved.
 /// These will be composed together with UnreseolvedVariant from fallback
 /// nodes (referenced with e.g. <linearGradient xlink:href="#fallback">) to form
@@ -164,6 +131,7 @@ enum UnresolvedVariant {
         r: Option<Length<Both>>,
         fx: Option<Length<Horizontal>>,
         fy: Option<Length<Vertical>>,
+        fr: Option<Length<Both>>,
     },
 }
 
@@ -183,6 +151,7 @@ enum Variant {
         r: Length<Both>,
         fx: Length<Horizontal>,
         fy: Length<Vertical>,
+        fr: Length<Both>,
     },
 }
 
@@ -198,12 +167,20 @@ impl UnresolvedVariant {
                 y2: y2.unwrap(),
             },
 
-            UnresolvedVariant::Radial { cx, cy, r, fx, fy } => Variant::Radial {
+            UnresolvedVariant::Radial {
+                cx,
+                cy,
+                r,
+                fx,
+                fy,
+                fr,
+            } => Variant::Radial {
                 cx: cx.unwrap(),
                 cy: cy.unwrap(),
                 r: r.unwrap(),
                 fx: fx.unwrap(),
                 fy: fy.unwrap(),
+                fr: fr.unwrap(),
             },
         }
     }
@@ -214,8 +191,20 @@ impl UnresolvedVariant {
                 x1.is_some() && y1.is_some() && x2.is_some() && y2.is_some()
             }
 
-            UnresolvedVariant::Radial { cx, cy, r, fx, fy } => {
-                cx.is_some() && cy.is_some() && r.is_some() && fx.is_some() && fy.is_some()
+            UnresolvedVariant::Radial {
+                cx,
+                cy,
+                r,
+                fx,
+                fy,
+                fr,
+            } => {
+                cx.is_some()
+                    && cy.is_some()
+                    && r.is_some()
+                    && fx.is_some()
+                    && fy.is_some()
+                    && fr.is_some()
             }
         }
     }
@@ -238,20 +227,29 @@ impl UnresolvedVariant {
             },
 
             (
-                UnresolvedVariant::Radial { cx, cy, r, fx, fy },
                 UnresolvedVariant::Radial {
-                    cx: fcx,
-                    cy: fcy,
-                    r: fr,
-                    fx: ffx,
-                    fy: ffy,
+                    cx,
+                    cy,
+                    r,
+                    fx,
+                    fy,
+                    fr,
+                },
+                UnresolvedVariant::Radial {
+                    cx: f_cx,
+                    cy: f_cy,
+                    r: f_r,
+                    fx: f_fx,
+                    fy: f_fy,
+                    fr: f_fr,
                 },
             ) => UnresolvedVariant::Radial {
-                cx: cx.or(fcx),
-                cy: cy.or(fcy),
-                r: r.or(fr),
-                fx: fx.or(ffx),
-                fy: fy.or(ffy),
+                cx: cx.or(f_cx),
+                cy: cy.or(f_cy),
+                r: r.or(f_r),
+                fx: fx.or(f_fx),
+                fy: fy.or(f_fy),
+                fr: fr.or(f_fr),
             },
 
             _ => *self, // If variants are of different types, then nothing to resolve
@@ -269,7 +267,14 @@ impl UnresolvedVariant {
                 y2: y2.or_else(|| Some(Length::<Vertical>::parse_str("0%").unwrap())),
             },
 
-            UnresolvedVariant::Radial { cx, cy, r, fx, fy } => {
+            UnresolvedVariant::Radial {
+                cx,
+                cy,
+                r,
+                fx,
+                fy,
+                fr,
+            } => {
                 let cx = cx.or_else(|| Some(Length::<Horizontal>::parse_str("50%").unwrap()));
                 let cy = cy.or_else(|| Some(Length::<Vertical>::parse_str("50%").unwrap()));
                 let r = r.or_else(|| Some(Length::<Both>::parse_str("50%").unwrap()));
@@ -277,8 +282,16 @@ impl UnresolvedVariant {
                 // fx and fy fall back to the presentational value of cx and cy
                 let fx = fx.or(cx);
                 let fy = fy.or(cy);
+                let fr = fr.or_else(|| Some(Length::<Both>::parse_str("0%").unwrap()));
 
-                UnresolvedVariant::Radial { cx, cy, r, fx, fy }
+                UnresolvedVariant::Radial {
+                    cx,
+                    cy,
+                    r,
+                    fx,
+                    fy,
+                    fr,
+                }
             }
         }
     }
@@ -299,16 +312,23 @@ impl Variant {
                 ))
             }
 
-            Variant::Radial { cx, cy, r, fx, fy } => {
+            Variant::Radial {
+                cx,
+                cy,
+                r,
+                fx,
+                fy,
+                fr,
+            } => {
                 let n_cx = cx.normalize(values, params);
                 let n_cy = cy.normalize(values, params);
                 let n_r = r.normalize(values, params);
                 let n_fx = fx.normalize(values, params);
                 let n_fy = fy.normalize(values, params);
-                let (new_fx, new_fy) = fix_focus_point(n_fx, n_fy, n_cx, n_cy, n_r);
+                let n_fr = fr.normalize(values, params);
 
                 cairo::Gradient::clone(&cairo::RadialGradient::new(
-                    new_fx, new_fy, 0.0, n_cx, n_cy, n_r,
+                    n_fx, n_fy, n_fr, n_cx, n_cy, n_r,
                 ))
             }
         }
@@ -348,6 +368,7 @@ pub struct RadialGradient {
     r: Option<Length<Both>>,
     fx: Option<Length<Horizontal>>,
     fy: Option<Length<Vertical>>,
+    fr: Option<Length<Both>>,
 }
 
 /// Main structure used during gradient resolution.  For unresolved
@@ -443,31 +464,27 @@ impl UnresolvedGradient {
 
     /// Looks for <stop> children inside a linearGradient or radialGradient node,
     /// and adds their info to the UnresolvedGradient &self.
-    fn add_color_stops_from_node(&mut self, node: &RsvgNode) {
-        let node_type = node.borrow().get_type();
+    fn add_color_stops_from_node(&mut self, node: &Node) {
+        assert!(
+            matches!(*node.borrow_element(), Element::LinearGradient(_) | Element::RadialGradient(_))
+        );
 
-        assert!(node_type == NodeType::LinearGradient || node_type == NodeType::RadialGradient);
+        for child in node.children().filter(|c| c.is_element()) {
+            let elt = child.borrow_element();
 
-        for child_node in node.children() {
-            let child = child_node.borrow();
+            if let Element::Stop(ref stop) = *elt {
+                if elt.is_in_error() {
+                    rsvg_log!("(not using gradient stop {} because it is in error)", child);
+                } else {
+                    let cascaded = CascadedValues::new_from_node(&child);
+                    let values = cascaded.get();
+                    let rgba = match values.stop_color() {
+                        StopColor(cssparser::Color::CurrentColor) => values.color().0,
+                        StopColor(cssparser::Color::RGBA(ref rgba)) => *rgba,
+                    };
 
-            if child.get_type() != NodeType::Stop {
-                continue;
-            }
-
-            let stop = child.get_impl::<Stop>();
-
-            if child.is_in_error() {
-                rsvg_log!("(not using gradient stop {} because it is in error)", child);
-            } else {
-                let cascaded = CascadedValues::new_from_node(&child_node);
-                let values = cascaded.get();
-                let rgba = match values.stop_color {
-                    StopColor(cssparser::Color::CurrentColor) => values.color.0,
-                    StopColor(cssparser::Color::RGBA(ref rgba)) => *rgba,
-                };
-
-                self.add_color_stop(stop.offset, rgba, values.stop_opacity.0);
+                    self.add_color_stop(stop.offset, rgba, values.stop_opacity().0);
+                }
             }
         }
     }
@@ -542,6 +559,7 @@ impl RadialGradient {
             r: self.r,
             fx: self.fx,
             fy: self.fy,
+            fr: self.fr,
         }
     }
 }
@@ -549,7 +567,7 @@ impl RadialGradient {
 macro_rules! impl_get_unresolved {
     ($gradient:ty) => {
         impl $gradient {
-            fn get_unresolved(&self, node: &RsvgNode) -> Unresolved {
+            fn get_unresolved(&self, node: &Node) -> Unresolved {
                 let mut gradient = UnresolvedGradient {
                     units: self.common.units,
                     transform: self.common.transform,
@@ -571,8 +589,8 @@ macro_rules! impl_get_unresolved {
 impl_get_unresolved!(LinearGradient);
 impl_get_unresolved!(RadialGradient);
 
-impl Common {
-    fn set_atts(&mut self, pbag: &PropertyBag<'_>) -> NodeResult {
+impl SetAttributes for Common {
+    fn set_attributes(&mut self, pbag: &PropertyBag<'_>) -> ElementResult {
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
                 expanded_name!("", "gradientUnits") => self.units = Some(attr.parse(value)?),
@@ -591,9 +609,9 @@ impl Common {
     }
 }
 
-impl NodeTrait for LinearGradient {
-    fn set_atts(&mut self, _: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
-        self.common.set_atts(pbag)?;
+impl SetAttributes for LinearGradient {
+    fn set_attributes(&mut self, pbag: &PropertyBag<'_>) -> ElementResult {
+        self.common.set_attributes(pbag)?;
 
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
@@ -610,19 +628,32 @@ impl NodeTrait for LinearGradient {
     }
 }
 
-impl NodeTrait for RadialGradient {
-    fn set_atts(&mut self, _: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
-        self.common.set_atts(pbag)?;
+impl Draw for LinearGradient {}
+
+impl SetAttributes for RadialGradient {
+    fn set_attributes(&mut self, pbag: &PropertyBag<'_>) -> ElementResult {
+        self.common.set_attributes(pbag)?;
+        // Create a local expanded name for "fr" because markup5ever doesn't have built-in
+        let expanded_name_fr = ExpandedName {
+            ns: &Namespace::from(""),
+            local: &LocalName::from("fr"),
+        };
 
         for (attr, value) in pbag.iter() {
-            match attr.expanded() {
-                expanded_name!("", "cx") => self.cx = Some(attr.parse(value)?),
-                expanded_name!("", "cy") => self.cy = Some(attr.parse(value)?),
-                expanded_name!("", "r") => self.r = Some(attr.parse(value)?),
-                expanded_name!("", "fx") => self.fx = Some(attr.parse(value)?),
-                expanded_name!("", "fy") => self.fy = Some(attr.parse(value)?),
+            let attr_expanded = attr.expanded();
 
-                _ => (),
+            if attr_expanded == expanded_name_fr {
+                self.fr = Some(attr.parse(value)?);
+            } else {
+                match attr_expanded {
+                    expanded_name!("", "cx") => self.cx = Some(attr.parse(value)?),
+                    expanded_name!("", "cy") => self.cy = Some(attr.parse(value)?),
+                    expanded_name!("", "r") => self.r = Some(attr.parse(value)?),
+                    expanded_name!("", "fx") => self.fx = Some(attr.parse(value)?),
+                    expanded_name!("", "fy") => self.fy = Some(attr.parse(value)?),
+
+                    _ => (),
+                }
             }
         }
 
@@ -630,14 +661,16 @@ impl NodeTrait for RadialGradient {
     }
 }
 
+impl Draw for RadialGradient {}
+
 macro_rules! impl_paint_source {
-    ($gradient:ty, $node_type:pat, $other_gradient:ty, $other_type:pat,) => {
-        impl PaintSource for $gradient {
+    ($gradient_type:ident, $other_type:ident) => {
+        impl PaintSource for $gradient_type {
             type Resolved = Gradient;
 
             fn resolve(
                 &self,
-                node: &RsvgNode,
+                node: &Node,
                 acquired_nodes: &mut AcquiredNodes,
             ) -> Result<Self::Resolved, AcquireError> {
                 let mut resolved = self.common.resolved.borrow_mut();
@@ -654,26 +687,17 @@ macro_rules! impl_paint_source {
 
                 while !gradient.is_resolved() {
                     if let Some(fragment) = fallback {
-                        let acquired = acquire_gradient(acquired_nodes, &fragment)?;
+                        let acquired = acquired_nodes.acquire(&fragment)?;
                         let acquired_node = acquired.get();
 
                         if stack.contains(acquired_node) {
                             return Err(AcquireError::CircularReference(acquired_node.clone()));
                         }
 
-                        let borrowed_node = acquired_node.borrow();
-                        let unresolved = match borrowed_node.get_type() {
-                            $node_type => {
-                                let a_gradient = borrowed_node.get_impl::<$gradient>();
-                                a_gradient.get_unresolved(&acquired_node)
-                            }
-
-                            $other_type => {
-                                let a_gradient = borrowed_node.get_impl::<$other_gradient>();
-                                a_gradient.get_unresolved(&acquired_node)
-                            }
-
-                            _ => unreachable!(),
+                        let unresolved = match *acquired_node.borrow_element() {
+                            Element::$gradient_type(ref g) => g.get_unresolved(&acquired_node),
+                            Element::$other_type(ref g) => g.get_unresolved(&acquired_node),
+                            _ => return Err(AcquireError::InvalidLinkType(fragment.clone())),
                         };
 
                         gradient = gradient.resolve_from_fallback(&unresolved.gradient);
@@ -696,19 +720,9 @@ macro_rules! impl_paint_source {
     };
 }
 
-impl_paint_source!(
-    LinearGradient,
-    NodeType::LinearGradient,
-    RadialGradient,
-    NodeType::RadialGradient,
-);
+impl_paint_source!(LinearGradient, RadialGradient);
 
-impl_paint_source!(
-    RadialGradient,
-    NodeType::RadialGradient,
-    LinearGradient,
-    NodeType::LinearGradient,
-);
+impl_paint_source!(RadialGradient, LinearGradient);
 
 impl AsPaintSource for Gradient {
     fn set_as_paint_source(
@@ -781,23 +795,12 @@ impl Gradient {
     }
 }
 
-/// Acquires a node of linearGradient or radialGradient type
-fn acquire_gradient(
-    acquired_nodes: &mut AcquiredNodes,
-    fragment: &Fragment,
-) -> Result<AcquiredNode, AcquireError> {
-    acquired_nodes.acquire(
-        fragment,
-        &[NodeType::LinearGradient, NodeType::RadialGradient],
-    )
-}
-
 #[cfg(test)]
 mod tests {
     use super::*;
-    use crate::float_eq_cairo::ApproxEqCairo;
-    use crate::node::{NodeData, NodeType, RsvgNode};
+    use crate::node::{Node, NodeData};
     use markup5ever::{namespace_url, ns, QualName};
+    use std::ptr;
 
     #[test]
     fn parses_spread_method() {
@@ -810,51 +813,26 @@ mod tests {
         assert!(SpreadMethod::parse_str("foobar").is_err());
     }
 
-    fn assert_tuples_equal(a: &(f64, f64), b: &(f64, f64)) {
-        assert_approx_eq_cairo!(a.0, b.0);
-        assert_approx_eq_cairo!(a.1, b.1);
-    }
-
-    #[test]
-    fn fixes_focus_point() {
-        // inside the circle
-        assert_tuples_equal(&fix_focus_point(1.0, 1.0, 2.0, 1.0, 3.0), &(1.0, 1.0));
-
-        // on the edge
-        assert_tuples_equal(&fix_focus_point(1.0, 1.0, 2.0, 1.0, 2.0), &(1.0, 1.0));
-
-        // outside the circle
-        assert_tuples_equal(&fix_focus_point(1.0, 1.0, 3.0, 1.0, 1.0), &(2.0, 1.0));
-    }
-
     #[test]
     fn gradient_resolved_from_defaults_is_really_resolved() {
-        let node = RsvgNode::new(NodeData::new(
-            NodeType::LinearGradient,
+        let bag = unsafe { PropertyBag::new_from_xml2_attributes(0, ptr::null()) };
+
+        let node = Node::new(NodeData::new_element(
             &QualName::new(None, ns!(svg), local_name!("linearGradient")),
-            None,
-            None,
-            Box::new(LinearGradient::default()),
+            &bag,
         ));
 
-        let borrow = node.borrow();
-        let g = borrow.get_impl::<LinearGradient>();
-        let Unresolved { gradient, .. } = g.get_unresolved(&node);
-        let gradient = gradient.resolve_from_defaults();
+        let unresolved = borrow_element_as!(node, LinearGradient).get_unresolved(&node);
+        let gradient = unresolved.gradient.resolve_from_defaults();
         assert!(gradient.is_resolved());
 
-        let node = RsvgNode::new(NodeData::new(
-            NodeType::RadialGradient,
+        let node = Node::new(NodeData::new_element(
             &QualName::new(None, ns!(svg), local_name!("radialGradient")),
-            None,
-            None,
-            Box::new(RadialGradient::default()),
+            &bag,
         ));
 
-        let borrow = node.borrow();
-        let g = borrow.get_impl::<RadialGradient>();
-        let Unresolved { gradient, .. } = g.get_unresolved(&node);
-        let gradient = gradient.resolve_from_defaults();
+        let unresolved = borrow_element_as!(node, RadialGradient).get_unresolved(&node);
+        let gradient = unresolved.gradient.resolve_from_defaults();
         assert!(gradient.is_resolved());
     }
 }
