@@ -26,8 +26,8 @@ make_property!(
             }
 
             fn compute(&self, v: &ComputedValues) -> Self {
-                let font_size = v.font_size.0.value();
-                let parent = v.baseline_shift.clone();
+                let font_size = v.font_size().0.value();
+                let parent = v.baseline_shift();
 
                 match (self.0.unit, parent.0.unit) {
                     (LengthUnit::Percent, _) => {
@@ -304,23 +304,13 @@ make_property!(
     "small-caps" => SmallCaps,
 );
 
-// https://drafts.csswg.org/css-fonts-4/#font-weight-prop
+// https://www.w3.org/TR/2008/REC-CSS2-20080411/fonts.html#propdef-font-weight
 make_property!(
     ComputedValues,
     FontWeight,
     default: FontWeightSpec::Normal,
+    inherits_automatically: true,
     newtype_parse: FontWeightSpec,
-    property_impl: {
-        impl Property<ComputedValues> for FontWeight {
-            fn inherits_automatically() -> bool {
-                true
-            }
-
-            fn compute(&self, v: &ComputedValues) -> Self {
-                FontWeight(self.0.compute(&v.font_weight.0))
-            }
-        }
-    }
 );
 
 // https://www.w3.org/TR/SVG/text.html#LetterSpacingProperty
